#include "stdafx.h"
#include "NScrollView.h"

#include "ScriptletConstant.h"

REGISTER_COMPONENT_CREATOR(SCRIPTLET_SCROLLVIEW_NODE, CNScrollView);
//////////////////////////////////////////////////////////////////////////
static const int SCROLLVIEW_WHEELHOP_DEF = -1;
static const int SCROLLVIEW_LINEHOP_DEF = 1;
static const int SCROLL_BTN_REPEAT_TIME = 100;

/************************************************************************/
/* CNScrollView */
/************************************************************************/
IMPLEMENT_RTTI(CNScrollView, CNPanel)

CNScrollView::CNScrollView()
{
	m_nWheelHop = SCROLLVIEW_WHEELHOP_DEF;
	m_nLineHop = SCROLLVIEW_LINEHOP_DEF;
	m_pFocusedComponent_Old = NULL;
}

CNScrollView::~CNScrollView()
{

}

void CNScrollView::OnCreate()
{	
	this->AddMouseListener(this);	
}

void CNScrollView::OnDestroy()
{
	this->RemoveMouseListener(this);
}
void CNScrollView::OnAddChild(CComponent* pChild)
{
	__super::OnAddChild(pChild);

	this->UpdateScrollInfo();
}
void CNScrollView::OnRemoveChild(CComponent* pChild)
{
	__super::OnRemoveChild(pChild);

	this->UpdateScrollInfo();
}
void CNScrollView::UpdateScrollInfo()
{
	CAtlList<CComponent*> lstChild;
	GetChildList(lstChild);

	CRect rcScroll;
	rcScroll.SetRectEmpty();

	POSITION pos = lstChild.GetHeadPosition();
	while (pos)
	{
		CComponent* pChild = lstChild.GetNext(pos);
		if ( NULL != pChild )
		{
			CRect rcChild;
			if (pChild->GetBounds(rcChild) == TRUE)
			{
				rcScroll.UnionRect(&rcScroll, &rcChild);
			}
		}
	}
	SCROLLINFO si;
	GetScrollInfo(&si, TRUE);
	si.nMax = rcScroll.Width();
	SetScrollInfo(&si, TRUE);

	GetScrollInfo(&si, FALSE);
	si.nMax = rcScroll.Height();
	SetScrollInfo(&si, FALSE);
}
void CNScrollView::SetScrollInfo( LPSCROLLINFO pSi, BOOL bHorz )
{
	if (pSi == NULL) return;
	if (m_pScrollBarHorz == NULL || m_pScrollBarVert == NULL) return;

	(bHorz) ?
	m_pScrollBarHorz->SetScrollInfo(pSi) :
	m_pScrollBarVert->SetScrollInfo(pSi);

	this->_Update_Layout_ScrollBars();
}

void CNScrollView::GetScrollInfo( LPSCROLLINFO pSi, BOOL bHorz )
{
	if (pSi == NULL) return;
	if (m_pScrollBarHorz == NULL || m_pScrollBarVert == NULL) return;

	(bHorz) ?
	m_pScrollBarHorz->GetScrollInfo(pSi) :
	m_pScrollBarVert->GetScrollInfo(pSi);
}

BOOL CNScrollView::GetClientRect( LPRECT lpRect )
{
	CRect rcClient;
	BOOL bResult = __super::GetClientRect(&rcClient);
	if (m_pScrollBarHorz != NULL && m_pScrollBarVert != NULL)
	{
		int nBtnWidthHorz = m_pScrollBarHorz->GetBtnWidth();

		int nBtnWidthVert = m_pScrollBarVert->GetBtnWidth();

		rcClient.bottom =	(m_pScrollBarHorz->IsVisible() == TRUE) ? 
							rcClient.bottom - nBtnWidthVert :
							rcClient.bottom;

		rcClient.right =	(m_pScrollBarVert->IsVisible() == TRUE) ?
							rcClient.right - nBtnWidthHorz :
							rcClient.right;

		::CopyRect(lpRect, &rcClient);
	}
	return bResult;
}
void CNScrollView::AddScrollListener( IScrollListener* pListener )
{
	m_pScrollBarHorz->AddScrollListener(pListener);
	m_pScrollBarVert->AddScrollListener(pListener);
}

void CNScrollView::RemoveScrollListener( IScrollListener* pListener )
{
	m_pScrollBarHorz->RemoveScrollListener(pListener);
	m_pScrollBarVert->RemoveScrollListener(pListener);
}
void CNScrollView::OnHScroll(CScrollEvent* pScrollEvent)
{
	CNScrollBar* pSbVert = pScrollEvent->m_pNScrollBar;
	ATLASSERT(IS_KIND_OF(CNScrollBar, pSbVert) == TRUE);
	SCROLLINFO si;
	pSbVert->GetScrollInfo(&si);

	if (pScrollEvent->m_nSBCode == SB_LINEUP)
	{
		si.nPos = max((int)si.nMin, (int)si.nPos - m_nLineHop);
	}
	else if (pScrollEvent->m_nSBCode == SB_LINEDOWN)
	{
		si.nPos = min((int)si.nMax - (int)si.nPage + 1, (int)si.nPos + m_nLineHop);
	}
	else if (pScrollEvent->m_nSBCode == SB_PAGEUP)
	{
		si.nPos = max((int)si.nMin, (int)si.nPos - (int)si.nPage);
	}
	else if (pScrollEvent->m_nSBCode == SB_PAGEDOWN)
	{
		si.nPos = min((int)si.nMax - (int)si.nPage + 1, (int)si.nPos + (int)si.nPage);
	}

	pSbVert->SetScrollInfo( &si );
	this->Invalidate();
}
void CNScrollView::OnVScroll(CScrollEvent* pScrollEvent)
{
	CNScrollBar* pSbVert = pScrollEvent->m_pNScrollBar;
	ATLASSERT(IS_KIND_OF(CNScrollBar, pSbVert) == TRUE);
	SCROLLINFO si;
	pSbVert->GetScrollInfo(&si);

	if (pScrollEvent->m_nSBCode == SB_LINEUP)
	{
		si.nPos = max((int)si.nMin, (int)si.nPos - m_nLineHop);
	}
	else if (pScrollEvent->m_nSBCode == SB_LINEDOWN)
	{
		si.nPos = min((int)si.nMax - (int)si.nPage + 1, (int)si.nPos + m_nLineHop);
	}
	else if (pScrollEvent->m_nSBCode == SB_PAGEUP)
	{
		si.nPos = max((int)si.nMin, (int)si.nPos - (int)si.nPage);
	}
	else if (pScrollEvent->m_nSBCode == SB_PAGEDOWN)
	{
		si.nPos = min((int)si.nMax - (int)si.nPage + 1, (int)si.nPos + (int)si.nPage);
	}

	pSbVert->SetScrollInfo( &si );
	this->Invalidate();
}
void CNScrollView::_Update_Layout_ScrollBars()
{
	if (m_pScrollBarHorz == NULL && m_pScrollBarVert == NULL) return;

	CRect rcClient;
	__super::GetClientRect(&rcClient);
	CRect rcDrawing;
	__super::GetDrawingBounds(&rcDrawing);
	
	CRect rcHorz;
	CRect rcVert;

	SCROLLINFO siHorz;
	m_pScrollBarHorz->GetScrollInfo(&siHorz);
	BOOL bScrollHorz = rcDrawing.Width() < siHorz.nMax;
	m_pScrollBarHorz->SetVisible(bScrollHorz);
	int nBtnWidthVert = m_pScrollBarVert->GetBtnWidth();

	SCROLLINFO siVert;
	m_pScrollBarVert->GetScrollInfo(&siVert);
	BOOL bScrollVert = rcDrawing.Height() < siVert.nMax;
	m_pScrollBarVert->SetVisible(bScrollVert);
	int nBtnWidthHorz = m_pScrollBarHorz->GetBtnWidth();
	
	siVert.nPage =	bScrollHorz ? 
					rcClient.Height() - nBtnWidthHorz : rcClient.Height();
	m_pScrollBarVert->SetScrollInfo(&siVert);

	siHorz.nPage =	bScrollVert ? 
					rcClient.Width() - nBtnWidthVert : rcClient.Width();
	m_pScrollBarHorz->SetScrollInfo(&siHorz);

	if (bScrollHorz == TRUE &&
		bScrollVert == TRUE)
	{
		rcHorz.SetRect
			(
			rcClient.left,
			rcClient.bottom - nBtnWidthHorz,
			rcClient.right - nBtnWidthVert,
			rcClient.bottom
			);
		m_pScrollBarHorz->MoveWindow(&rcHorz);

		rcVert.SetRect
			(
			rcClient.right - nBtnWidthVert,
			rcClient.top,
			rcClient.right,
			rcClient.bottom - nBtnWidthHorz
			);
		m_pScrollBarVert->MoveWindow(&rcVert);
	}
	else if (	bScrollHorz == TRUE &&
		bScrollVert == FALSE)
	{
		rcHorz.SetRect
			(
			rcClient.left,
			rcClient.bottom - nBtnWidthHorz,
			rcClient.right,
			rcClient.bottom
			);
		m_pScrollBarHorz->MoveWindow(&rcHorz);
	}
	else if (	bScrollHorz == FALSE &&
		bScrollVert == TRUE)
	{
		rcVert.SetRect
			(
			rcClient.right - nBtnWidthVert,
			rcClient.top,
			rcClient.right,
			rcClient.bottom
			);
		m_pScrollBarVert->MoveWindow(&rcVert);
	}
}
void CNScrollView::OnSize(INT32 nWidth, INT32 nHeight)
{
	__super::OnSize(nWidth, nHeight);

	UpdateScrollInfo();
	_Update_Layout_ScrollBars();
}
BOOL CNScrollView::_GetScrollPos( LPPOINT ppt )
{
	ATLASSERT(ppt != NULL); //hhdatlassert
	if ( !(ppt != NULL) ) return FALSE;
	if (m_pScrollBarHorz==NULL || m_pScrollBarVert==NULL) return FALSE;

	SCROLLINFO siHorz;
	m_pScrollBarHorz->GetScrollInfo(&siHorz);
	ppt->x = siHorz.nPos;

	SCROLLINFO siVert;
	m_pScrollBarVert->GetScrollInfo(&siVert);
	ppt->y = siVert.nPos;

	return TRUE;
}

void CNScrollView::SetWheelHop( int nWheelHop )
{
	ATLASSERT(nWheelHop >= 0); //hhdatlassert
	if ( !(nWheelHop >= 0) ) return;

	m_nWheelHop = nWheelHop;
}

int CNScrollView::GetWheelHop()
{
	return m_nWheelHop;
}
void CNScrollView::SetLineHop(int nLineHop)
{
	ATLASSERT(nLineHop >= 0); //hhdatlassert
	if ( !(nLineHop >= 0) ) return;

	m_nLineHop = nLineHop;
}
int CNScrollView::GetLineHop()
{
	return m_nLineHop;
}
void CNScrollView::OnMouseEntered( CMouseEvent* pMouseEvent )
{
	if (this->GetFocus() != this)
	{
		m_pFocusedComponent_Old = this->GetFocus();
		this->SetFocus();
	}
}

void CNScrollView::OnMouseExited( CMouseEvent* pMouseEvent )
{
	if (this->GetFocus() != m_pFocusedComponent_Old &&
		m_pFocusedComponent_Old != NULL)
	{
		m_pFocusedComponent_Old->SetFocus();
	}
}

IMPLEMENT_MESSAGE_HANDLER(CNScrollView, OnMouseWheel)
{
	bHandled = FALSE;
	BOOL bVScroll = m_pScrollBarVert->IsVisible();
	
	POINT ptMouse;
	ptMouse.x = GET_X_LPARAM(lParam); 
	ptMouse.y = GET_Y_LPARAM(lParam); 
	CRect rcWnd;
	this->GetWindowRect(&rcWnd);
	BOOL bPtInRect = rcWnd.PtInRect(ptMouse);
	
	if (bVScroll == TRUE &&
		bPtInRect == TRUE)
	{
		int zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
		int nScrollHop = 0;
		
		if (m_nWheelHop != SCROLLVIEW_WHEELHOP_DEF)
		{
			nScrollHop = m_nWheelHop;
		}
		else
		{
			nScrollHop = abs(zDelta);
		}
		
		SCROLLINFO siVert;
		m_pScrollBarVert->GetScrollInfo(&siVert);

		if (zDelta < 0)
		{
			siVert.nPos = min(siVert.nMax - siVert.nPage + 1, siVert.nPos + nScrollHop);
			m_pScrollBarVert->SetScrollInfo(&siVert);
			this->Invalidate();
		}
		else if (zDelta > 0)
		{
			siVert.nPos = max(siVert.nMin, siVert.nPos - nScrollHop);
			m_pScrollBarVert->SetScrollInfo(&siVert);
			this->Invalidate();
		}
	}
	return 0;
}

void CNScrollView::AddScrollBar( CScrollBarUI_Img::PSCROLLBARUI_INFO pInfoHorz, CScrollBarUI_Img::PSCROLLBARUI_INFO pInfoVert )
{
	if (m_pScrollBarHorz != NULL || m_pScrollBarVert != NULL) return;
	if (pInfoHorz != NULL)
	{
		m_pScrollBarUI_Img_Horz.Attach(new CScrollBarUI_Img(pInfoHorz));
		m_pScrollBarHorz.Attach
			(
			new CNScrollBar
			(
			m_pScrollBarUI_Img_Horz.m_p,
			NSB_STYLE_HSCROLL, 
			m_pScrollBarUI_Img_Horz->GetBtnWidth()	
			)	
			);
	}
	else
	{
		m_pScrollBarHorz.Attach(new CNScrollBar(NULL, NSB_STYLE_HSCROLL));
	}
	
	if (m_pScrollBarVert == NULL)
	{
		if (pInfoVert != NULL)
		{
			m_pScrollBarUI_Img_Vert.Attach(new CScrollBarUI_Img(pInfoVert));
			m_pScrollBarVert.Attach
				(
				new CNScrollBar
				(
				m_pScrollBarUI_Img_Vert.m_p,
				NSB_STYLE_VSCROLL,
				m_pScrollBarUI_Img_Vert->GetBtnWidth()
				)
				);
		}
		else
		{
			m_pScrollBarVert.Attach(new CNScrollBar(NULL, NSB_STYLE_VSCROLL));
		}
	}
	SCROLLINFO si = {0,};
	m_pScrollBarHorz->SetVisible(FALSE);
	m_pScrollBarHorz->SetScrollInfo(&si);
	m_pScrollBarHorz->SetTimerPeriod(SCROLL_BTN_REPEAT_TIME);
	AddNCComponent(m_pScrollBarHorz.m_p);

	m_pScrollBarVert->SetVisible(FALSE);
	m_pScrollBarVert->SetScrollInfo(&si);
	m_pScrollBarVert->SetTimerPeriod(SCROLL_BTN_REPEAT_TIME);
	AddNCComponent(m_pScrollBarVert.m_p);

	this->AddScrollListener(this);
}